use super::errors::BlackJackError;
use super::hand::Hand;
use crate::cards::StandardCard as Card;
use crate::deck::Deck;
use failure::Error;
use std::cmp::PartialEq;

/// A struct to represent the games state
#[derive(Debug, Clone)]
pub enum GameState {
    /// Game is currently in progress
    InProgress,
    /// Player has won the game
    PlayerWon,
    /// Player has lost the game
    PlayerLost,
    /// Game is frozen
    Frozen {
        /// Player's cards
        player: Vec<Card>,
        /// Dealer's cards
        dealer: Vec<Card>,
        /// The remaining deck
        deck: Vec<Card>,
        /// Has the player chosen to stay
        player_stay: bool,
        /// Has the dealer chosen to stay
        dealer_stay: bool,
        /// Is this the first turn
        first_turn: bool,
    },
}

impl PartialEq for GameState {
    fn eq(&self, other: &GameState) -> bool {
        match self {
            GameState::InProgress => match other {
                GameState::InProgress => true,
                _ => false,
            },
            GameState::PlayerWon => match other {
                GameState::PlayerWon => true,
                _ => false,
            },
            GameState::PlayerLost => match other {
                GameState::PlayerLost => true,
                _ => false,
            },
            GameState::Frozen { .. } => match other {
                GameState::Frozen { .. } => true,
                _ => false,
            },
        }
    }
}

// TODO: impl `display_game(&self) -> [(u8,Vec<Cards>),(u8,Vec<Cards>);2]` for BlackJack

/// BlackJack Game
#[derive(Debug, Clone)]
pub struct BlackJack {
    /// Players Hand
    player: Hand,
    /// Dealers Hand
    dealer: Hand,
    /// First Turn, used for display purposes
    first_turn: bool,
    /// Deck of cards in the game
    deck: Deck<crate::cards::StandardCard>,
    /// Has the player chosen to stay
    player_stay_status: bool,
    /// Has the dealer chosen to stay
    dealer_stay_status: bool,
}

impl BlackJack {
    /// Create a new `Blackjack` Game
    pub fn new<T: Into<Option<usize>>>(max_decks: T) -> Self {
        // Create a new deck
        let mut deck = Deck::new(max_decks, true);
        // Create The players hand
        let player = Hand {
            cards: deck.draw_many(2),
        };

        // Create the dealers hand
        let dealer = Hand {
            cards: deck.draw_many(2),
        };

        Self {
            player,
            dealer,
            deck,
            dealer_stay_status: false,
            player_stay_status: false,
            first_turn: true,
        }
    }
    /// Freeze the `BlackJack` Game
    /// GameState::PlayerWon - Player has one, no other data needed
    /// GameState::PlayerLost - Player has lost, no other data needed
    /// GameState::Frozen - Game is in progress, can be restored later
    pub fn freeze(self) -> GameState {
        match self.status() {
            GameState::InProgress => GameState::Frozen {
                player: self.player.cards,
                dealer: self.dealer.cards,
                deck: self.deck.cards,
                first_turn: self.first_turn,
                player_stay: self.player_stay_status,
                dealer_stay: self.dealer_stay_status,
            },
            gs => gs,
        }
    }
    /// Restore a `BlackKJack` game from `GameState::Frozen`
    pub fn dethaw(g: GameState) -> Option<Self> {
        match g {
            GameState::Frozen {
                player,
                dealer,
                deck,
                first_turn,
                player_stay,
                dealer_stay,
            } => Some(Self {
                player: Hand { cards: player },
                dealer: Hand { cards: dealer },
                deck: Deck {
                    cards: deck,
                    random_draw: true,
                },
                first_turn,
                player_stay_status: player_stay,
                dealer_stay_status: dealer_stay,
            }),
            _ => None,
        }
    }
    /// Return the `GameState`
    /// Possible values:
    /// `GameState::InProgress` - Game is in progress
    /// `GameState::PlayerWon` - Player has won
    /// `GameState::PlayerLost` - Player has lost
    pub fn status(&self) -> GameState {
        // Question: What if the first turn is 21?
        // TODO: Rework logic to solve above question
        if self.first_turn {
            return GameState::InProgress;
        }
        // Calculate the scores
        let player_score = self.player.score();
        let dealer_score = self.dealer.score();

        // If the player has 5 cards and less than 21 thats a win
        if self.player.cards.len() == 5 && player_score <= 21 {
            return GameState::PlayerWon;
        }

        // Same applies for the dealer
        if self.dealer.cards.len() == 5 && dealer_score <= 21 {
            return GameState::PlayerWon;
        }

        // 21 Is a win
        if player_score == 21 {
            return GameState::PlayerWon;
        }

        // Or a loss if the dealer has it
        if dealer_score == 21 {
            return GameState::PlayerLost;
        }

        // Over 21 is a loss
        if player_score > 21 {
            return GameState::PlayerLost;
        }

        if !(self.player_stay_status || self.dealer_stay_status) {
            return GameState::InProgress;
        }

        // A draw is a loss
        if player_score == dealer_score {
            return GameState::PlayerLost;
        }

        // Over 21 is a loss
        if dealer_score > 21 {
            return GameState::PlayerWon;
        }

        // Highest score decides win
        if player_score > dealer_score {
            return GameState::PlayerWon;
        }
        if player_score < dealer_score {
            return GameState::PlayerLost;
        }

        // Should be unreachable, but just in case
        GameState::InProgress
    }
    /// Draws a card if game is InProgress, will error otherwise
    pub fn player_hit(&mut self) -> Result<(), Error> {
        // Check the game status
        match self.status() {
            GameState::InProgress => {
                self.first_turn = false;
                self.player.add_card(
                    self.deck
                        .draw()
                        .expect("Game should not reach a point of being out of cards"),
                );
                Ok(())
            }
            GameState::PlayerLost => Err(BlackJackError::PlayerLost)?,
            GameState::PlayerWon => Err(BlackJackError::PlayerWon)?,
            // Frozen is unreachable
            _ => unreachable!(),
        }
    }
    /// Finish the blackjack game (player choses to stay)
    /// The dealer will draw cards as necessary
    /// game result will be returned
    pub fn finish(mut self) -> GameState {
        self.player_stay_status = true;
        self.dealer_logic();
        match self.status() {
            // Game state should only be Win/Lose
            GameState::InProgress | GameState::Frozen { .. } => unreachable!(),
            gs => gs,
        }
    }
    /// Handles the dealers logic for `BlackJack::finish`
    fn dealer_logic(&mut self) {
        self.first_turn = false;
        // Dealer stops attempting at 17 / Higher
        // And action is only taken if the player didnt already win/lose
        while self.status() == GameState::InProgress && self.dealer.score() <= 17 {
            self.dealer.add_card(
                self.deck
                    .draw()
                    .expect("Game should not reach a point where the game is out of cards"),
            )
        }
        self.dealer_stay_status = true;
    }
}
