use std::fmt;
/// Error raised by Coin Toss
#[derive(Debug, Fail, Copy, Clone)]
pub enum CoinError {
    /// Failed to parse coin side
    #[fail(display = "Not a valid side")]
    InvalidSide,
}

/// Enum representing a coin
#[derive(Clone, Copy, Debug, Eq, Hash, PartialEq, PartialOrd, Ord)]
pub enum Coin {
    /// Head side of the coin
    Heads,
    /// Tails side of the coin
    Tails,
}

impl Coin {
    /// Flips a coin, returning a Coin with its current Side up
    pub fn flip() -> Coin {
        Coin::from(rand::random::<bool>())
    }
}

impl fmt::Display for Coin {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(
            f,
            "{}",
            match self {
                Coin::Heads => "Heads",
                Coin::Tails => "Tails",
            }
        )
    }
}

/// Coin toss struct
#[derive(Copy, Clone, Debug)]
pub struct CoinToss {
    /// The coin's state that the player guessed
    pub guess: Coin,
    /// The coin's state that happened after the toss
    pub real: Coin,
}

impl std::convert::Into<bool> for CoinToss {
    fn into(self) -> bool {
        self.guess == self.real
    }
}

impl std::convert::From<bool> for Coin {
    fn from(b: bool) -> Coin {
        if b {
            Coin::Heads
        } else {
            Coin::Tails
        }
    }
}

impl fmt::Display for CoinToss {
    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
        write!(f, "Guess: {}\nResult: {}", self.guess, self.real)
    }
}

impl CoinToss {
    /// Make a guess on what side the coin will land on
    pub fn guess(coin_side: &str) -> Result<Self, CoinError> {
        if !["heads", "h", "tails", "t"].contains(&coin_side) {
            Err(CoinError::InvalidSide)
        } else {
            Ok(CoinToss {
                guess: Coin::from(coin_side.starts_with('h')),
                real: Coin::flip(),
            })
        }
    }
    /// Make a guess on what side the coin will land on
    pub fn guess_coin(guess: Coin) -> Self {
        CoinToss {
            guess,
            real: Coin::flip(),
        }
    }
}
