use crate::cards::StandardCard;
#[allow(deprecated)] // Deprecated for external use
use crate::deck_of_cards::STANDARD_DECK_OF_CARDS;
use rand::{seq::SliceRandom, thread_rng, Rng};
use std::error::Error as StdError;
use std::fmt::{Display, Formatter, Result as FmtResult};

/// Standard card deck
#[derive(Clone, Debug)]
pub struct Deck<C: Clone + Copy + std::fmt::Debug> {
    /// Deck of cards
    pub cards: Vec<C>,
    /// If true, StandardDeck::draw will draw randomly
    /// This improves performance in games that wont be going through the whole deck
    pub random_draw: bool,
}

/// Deck error
#[derive(Debug, Clone, Copy)]
pub enum DeckError {
    /// No cards left in the deck
    NoCard,
}

impl Display for DeckError {
    fn fmt(&self, f: &mut Formatter) -> FmtResult {
        f.write_str(self.description())
    }
}

impl StdError for DeckError {
    fn description(&self) -> &str {
        match *self {
            DeckError::NoCard => "No Cards are left in the deck",
        }
    }
}

impl<C: Clone + Copy + std::fmt::Debug> Deck<C> {
    /// Draw a card from the deck at random
    pub fn draw(&mut self) -> Result<C, DeckError> {
        if self.cards.is_empty() {
            return Err(DeckError::NoCard);
        }
        if !self.random_draw {
            return Ok(self.cards.remove(0));
        }
        let len = self.cards.len();
        let index = thread_rng().gen_range(0, len);
        Ok(self.cards.remove(index))
    }
    /// Draw up to <count> cards, will stop early if out of cards
    pub fn draw_many(&mut self, count: usize) -> Vec<C> {
        if self.cards.is_empty() || count == 0 {
            return Vec::with_capacity(0);
        }
        if self.random_draw {
            let (shuffled, not_shuffled) = self.cards.partial_shuffle(&mut thread_rng(), count);
            let shuffled = shuffled.into();
            let not_shuffled = not_shuffled.to_vec();
            self.cards = not_shuffled;

            return shuffled;
        }
        let len = self.cards.len();
        if count >= len {
            self.cards.drain(..).collect()
        } else {
            self.cards.split_off(len - count)
        }
    }
    /// Insert a card into the deck
    pub fn insert(&mut self, card: C) {
        self.cards.push(card);
    }
    /// Convert the dec into a vector of cards
    pub fn into_vec(self) -> Vec<C> {
        self.cards
    }
    /// Shuffle the deck
    pub fn shuffle(&mut self) {
        self.cards.shuffle(&mut thread_rng());
    }
}

impl Deck<StandardCard> {
    /// Create a new deck
    /// sets: The number of sets to create
    /// None is the fastest way to create a deck size of 1
    /// Otherwise just pass an usize
    pub fn new<T: Into<Option<usize>>>(sets: T, random_draw: bool) -> Self {
        #[allow(deprecated)]
        let cards: Vec<StandardCard> = match sets.into() {
            None => STANDARD_DECK_OF_CARDS.to_vec(),
            Some(sets) => {
                let mut vec = Vec::with_capacity(52 * sets);
                for _ in 0..sets {
                    vec.extend(STANDARD_DECK_OF_CARDS.iter())
                }
                vec
            }
        };
        Self {
            cards: cards,
            random_draw,
        }
    }
}

impl Default for Deck<StandardCard> {
    #[allow(deprecated)]
    fn default() -> Deck<StandardCard> {
        Self {
            cards: STANDARD_DECK_OF_CARDS.to_vec(),
            random_draw: true,
        }
    }
}
